# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinForms library

    @copyright: 2012, 2013 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from compiler import parse
from compiler.ast import Const, Dict, Discard, List, Module, Stmt
from MoinMoin.action import do_show
from MoinMoin.Page import Page
from MoinMoin import security, wikiutil
from MoinSupport import *
import re

__version__ = "0.1"

form_field_regexp_str = r"<<Form(Field|Message)\((.*?)\)>>"
form_field_regexp = re.compile(form_field_regexp_str, re.DOTALL)

# Common action functionality.

class MoinFormHandlerAction:

    "A handler action that can be specialised for individual forms."

    def __init__(self, pagename, request):
        self.pagename = pagename
        self.request = request
        self.access_handler = None

    def getAccessHandler(self, attributes):
        return FormAccess(self.pagename, self.request, attributes)

    def processForm(self):

        """
        Interpret the request details and modify them according to the structure
        of the interpreted information.
        """

        _ = self.request.getText

        # Get the form fields and obtain the hierarchical field structure.

        form = get_form(self.request)
        fields = getFields(form, remove=True)

        # Modify and validate the form.

        self.modifyFields(fields)

        # Get the form definition.

        attributes, text = self.getFormForFragment(fields)
        structure = getFormStructure(text, self.request)
        self.access_handler = self.getAccessHandler(attributes)

        # Check the permissions on the form.

        if not self.checkPermissions("write"):
            self.request.theme.add_msg(_("You do not appear to have access to this form."), "error")
            do_show(self.pagename, self.request)
            return

        # Without any form definition, the page is probably the wrong one.

        if not structure:
            self.request.theme.add_msg(_("This page does not provide a form."), "error")
            do_show(self.pagename, self.request)
            return

        # With a form definition, attempt to validate the fields.

        if self.validateFields(fields, structure):
            self.finished(fields, form)
        else:
            self.unfinished(fields, form)

    def finished(self, fields, form):

        "Handle the finished 'fields' and 'form'."

        self.storeFields(fields)
        self.unfinished(fields, form)

    def unfinished(self, fields, form):

        "Handle the unfinished 'fields' and 'form'."

        # Serialise and show the form.

        self.serialiseFields(fields, form)
        do_show(self.pagename, self.request)

    def getFormForFragment(self, fields):

        "Return the attributes and text of the form being handled."

        fragment = fields.get("fragment", [None])[0]
        text = Page(self.request, self.pagename).get_raw_body()
        return getFormForFragment(text, fragment)

    def checkPermissions(self, action):

        """
        Check the permissions of the user against any restrictions specified in
        the form's 'attributes'.
        """

        return self.access_handler.checkPermissions(action)

    def validateFields(self, fields, structure):

        """
        Validate the given 'fields' using the given form 'structure',
        introducing error fields where the individual fields do not conform to
        their descriptions.
        """

        return self.validateFieldsUsingStructure(fields, structure)

    def validateFieldsUsingStructure(self, fields, structure):

        "Validate the given 'fields' using the given 'structure'."

        _ = self.request.getText
        valid = True

        for key, definition in structure.items():
            value = fields.get(key)

            # Enter form sections and validate them.

            if isinstance(definition, dict):
                if value:
                    for element in getSectionElements(value):
                        valid = self.validateFieldsUsingStructure(element, structure[key]) and valid

            # Validate individual fields.

            elif structure.has_key(key):
                path, dictpage, label, section, field_args, allowed_values = definition
                errors = []

                # Test for obligatory values.

                if not value or not value[0]:
                    if field_args.get("required"):

                        # Detect new parts of the structure and avoid producing
                        # premature error messages.

                        if not fields.has_key("_new"):
                            errors.append(_("This field must be filled out."))
                        else:
                            valid = False
                else:
                    # Test for unacceptable values.

                    if allowed_values and set(value).difference(allowed_values):
                        errors.append(_("At least one of the choices is not acceptable."))

                    # Test the number of values.

                    if field_args.get("type") == "select":
                        if field_args.has_key("maxselected"):
                            if len(value) > int(field_args["maxselected"]):
                                errors.append(_("Incorrect number of choices given: need %s.") % field_args["maxselected"])

                if errors:
                    fields["%s-error" % key] = errors
                    valid = False

        return valid

    def serialiseFields(self, fields, form, path=None):

        """
        Serialise the given 'fields' to the given 'form', using the given 'path'
        to name the entries.
        """

        for key, value in fields.items():

            # Serialise sections.

            if isinstance(value, dict):
                for index, element in enumerate(getSectionElements(value)):
                    element_ref = "%s$%s" % (key, index)

                    self.serialiseFields(element, form,
                        path and ("%s/%s" % (path, element_ref)) or element_ref
                        )

            # Serialise fields.

            else:
                form[path and ("%s/%s" % (path, key)) or key] = value

    def modifyFields(self, fields):

        "Modify the given 'fields', removing and adding items."

        # First, remove fields.

        for key in fields.keys():
            if key.startswith("_remove="):
                self.removeField(key[8:], fields)

        # Then, add fields.

        for key in fields.keys():
            if key.startswith("_add="):
                self.addField(key[5:], fields)

    def removeField(self, path, fields):

        """
        Remove the section element indicated by the given 'path' from the
        'fields'.
        """

        section, (name, index) = getSectionForPath(path, fields)
        try:
            del section[name][index]
        except KeyError:
            pass

    def addField(self, path, fields):

        """
        Add a section element indicated by the given 'path' to the 'fields'.
        """

        section, (name, index) = getSectionForPath(path, fields)
        placeholder = {"_new" : ""}

        if section.has_key(name):
            indexes = section[name].keys()
            max_index = max(map(int, indexes))
            section[name][max_index + 1] = placeholder
        else:
            max_index = -1
            section[name] = {0 : placeholder}

    # Storage of form submissions.

    def storeFields(self, fields):

        """
        Store the given 'fields' as a Python object representation, subject to
        the given 'attributes'.
        """

        store = FormStore(self.access_handler)
        store.append(repr(fields))

    def loadFields(self, number):

        "Load the fields associated with the given submission 'number'."

        store = FormStore(self.access_handler)
        return loadFields(store, number)

def loadFieldsForRequest(number, attrs, request):

    """
    Load the fields corresponding to the given record 'number', using the form
    'attrs' to control access to the form data, and using the 'request' for
    access and retrieval purposes.
    """

    access_handler = FormAccess(request.page.page_name, request, attrs)
    store = FormStore(access_handler)
    return loadFields(store, number)

def loadFields(store, number):

    """
    Load the fields from the 'store' that are associated with the given
    submission 'number'.
    """

    try:
        module = parse(store[number])
        if checkStoredFormData(module):
            return eval(module)

    # NOTE: Should indicate any errors in retrieving form data.

    except:
        pass

    return {}

def checkStoredFormData(node):

    """
    Check the syntax 'node' and its descendants for suitability as parts of
    a field definition.
    """

    for child in node.getChildNodes():
        if isinstance(child, Const):
            pass
        elif not isinstance(child, (Dict, Discard, List, Module, Stmt)) or not checkStoredFormData(child):
            return False

    return True

class FormAccess:

    "A means of checking access to form data."

    def __init__(self, pagename, request, attributes):
        self.pagename = pagename
        self.request = request
        self.attributes = attributes

    def checkPermissions(self, action):

        """
        Check the permissions of the user against any restrictions specified in
        the form's 'attributes'.
        """

        user = self.request.user

        # Use the access definition if one is given.

        if self.attributes.has_key("access"):
            access = self.attributes["access"]
            acl = security.AccessControlList(self.request.cfg, [access])
            policy = lambda request, pagename, username, action: acl.may(request, username, action)

        # Otherwise, use the page permissions.

        else:
            policy = security._check

        # The "read" action is only satisfied by the "admin" role.

        return user and (
            action != "read" and policy(self.request, self.pagename, user.name, action) or
            action == "read" and policy(self.request, self.pagename, user.name, "admin")
            )

class FormStore(ItemStore):

    "A form-specific storage mechanism."

    def __init__(self, handler):

        "Initialise the store with the form 'handler'."

        self.handler = handler
        page = Page(handler.request, handler.pagename)
        ItemStore.__init__(self, page, "forms", "form-locks")

    def can_write(self):

        """
        Permit writing of form data using the form attributes or page
        permissions.
        """

        return self.handler.checkPermissions("write")

    def can_read(self):

        """
        Permit reading of form data using the form attributes or page
        permissions.
        """

        return self.handler.checkPermissions("read")

# Form and field information.

def getFormStructure(text, request, path=None, structure=None):

    """
    For the given form 'text' and using the 'request', return details of the
    form for the section at the given 'path' (or the entire form if 'path' is
    omitted), populating the given 'structure' (or populating a new structure if
    'structure' is omitted).
    """

    if structure is None:
        structure = {}

    for format, attributes, body in getFragments(text, True):

        # Get field details at the current level.

        if format is None:
            structure.update(getFormFields(body, path, request))

        # Where a section is found, get details from within the section.

        elif format == "form":
            if attributes.has_key("section"):
                section_name = attributes["section"]
                section = structure[section_name] = {}
                getFormStructure(body, request, path and ("%s/%s" % (path, section_name)) or section_name, section)
            elif attributes.has_key("message"):
                getFormStructure(body, request, path, structure)
            elif attributes.has_key("not-message"):
                getFormStructure(body, request, path, structure)

        # Get field details from other kinds of region.

        elif format != "form":
            getFormStructure(body, request, path, structure)

    return structure

def getFormForFragment(text, fragment=None):

    """
    Return the form region from the given 'text' for the specified 'fragment'.
    If no fragment is specified, the first form region is returned. The form
    region is described using a tuple containing the attributes for the form
    and the body text of the form.
    """

    for format, attributes, body in getFragments(text):
        if format == "form" and (not fragment or attributes.get("fragment") == fragment):
            return attributes, body

    return {}, None

def getFieldArguments(field_definition):

    "Return the parsed arguments from the given 'field_definition' string."

    field_args = {}

    for field_arg in field_definition.split():
        if field_arg == "required":
            field_args[field_arg] = True
            continue

        # Record the key-value details.

        try:
            argname, argvalue = field_arg.split("=", 1)
            field_args[argname] = argvalue

        # Single keywords are interpreted as type descriptions.

        except ValueError:
            if not field_args.has_key("type"):
                field_args["type"] = field_arg

    return field_args

# Common formatting functions.

def getFormOutput(text, fields, path=None, fragment=None, repeating=None, index=None):

    """
    Combine regions found in the given 'text' and then return them as a single
    block. The reason for doing this, as opposed to just passing each region to
    a suitable parser for formatting, is that form sections may break up
    regions, and such sections may not define separate subregions but instead
    act as a means of conditional inclusion of text into an outer region.

    The given 'fields' are used to populate fields provided in forms and to
    control whether sections are populated or not.

    The optional 'path' is used to adjust form fields to refer to the correct
    part of the form hierarchy.

    The optional 'fragment' is used to indicate the form to which the fields
    belong.

    The optional 'repeating' and 'index' is used to refer to individual values
    of a designated field.
    """

    output = []
    section = fields

    for region in getRegions(text, True):
        format, attributes, body, header, close = getFragmentFromRegion(region)

        # Adjust any FormField macros to use hierarchical names.

        if format is None:
            output.append((path or fragment or repeating) and
                adjustFormFields(body, path, fragment, repeating, index) or body)

        # Include form sections only if fields exist for those sections.

        elif format == "form":
            section_name = attributes.get("section")
            message_name = attributes.get("message")
            absent_message_name = attributes.get("not-message")

            # Sections are groups of fields in their own namespace.

            if section_name and section.has_key(section_name):

                # Iterate over the section contents ignoring the given indexes.

                for index, element in enumerate(getSectionElements(section[section_name])):
                    element_ref = "%s$%s" % (section_name, index)

                    # Get the output for the section.

                    output.append(getFormOutput(body, element,
                        path and ("%s/%s" % (path, element_ref)) or element_ref, fragment))

            # Message regions are conditional on a particular field and
            # reference the current namespace.

            elif message_name and section.has_key(message_name):

                if attributes.get("repeating"):
                    for index in range(0, len(section[message_name])):
                        output.append(getFormOutput(body, section, path, fragment, message_name, index))
                else:
                    output.append(getFormOutput(body, section, path, fragment))

            # Not-message regions are conditional on a particular field being
            # absent. They reference the current namespace.

            elif absent_message_name and not section.has_key(absent_message_name):
                output.append(getFormOutput(body, section, path, fragment))

        # Inspect and include other regions.

        else:
            output.append(header)
            output.append(getFormOutput(body, section, path, fragment, repeating, index))
            output.append(close)

    return "".join(output)

def getFormFields(body, path, request):

    "Return a dictionary of fields from the given 'body' at the given 'path'."

    fields = {}
    cache = {}
    type = None

    for i, match in enumerate(form_field_regexp.split(body)):
        state = i % 3

        if state == 1:
            type = match
        elif state == 2 and type == "Field":
            args = {}

            # Obtain the macro arguments, adjusted to consider the path.

            name, path, dictpage, label, section, fragment = \
                getMacroArguments(adjustMacroArguments(parseMacroArguments(match), path))

            # Obtain field information from the cache, if possible.

            cache_key = (name, dictpage)

            if cache.has_key(cache_key):
                field_args, allowed_values = cache[cache_key]

            # Otherwise, obtain field information from any WikiDict.

            else:
                field_args = {}
                allowed_values = None

                if dictpage:
                    wikidict = getWikiDict(dictpage, request)
                    if wikidict:
                        field_definition = wikidict.get(name)
                        if field_definition:
                            field_args = getFieldArguments(field_definition)
                            if field_args.has_key("source"):
                                sourcedict = getWikiDict(field_args["source"], request)
                                if sourcedict:
                                    allowed_values = sourcedict.keys()

                cache[cache_key] = field_args, allowed_values

            # Store the field information.

            fields[name] = path, dictpage, label, section, field_args, allowed_values

    return fields

def adjustFormFields(body, path, fragment, repeating=None, index=None):

    """
    Return a version of the 'body' with the names in FormField macros updated to
    incorporate the given 'path' and 'fragment'. If 'repeating' is specified,
    any field with such a name will be adjusted to reference the value with the
    given 'index'.
    """

    result = []
    type = None

    for i, match in enumerate(form_field_regexp.split(body)):
        state = i % 3

        # Reproduce normal text as is.

        if state == 0:
            result.append(match)

        # Capture the macro type.

        elif state == 1:
            type = match

        # Substitute the macro and modified arguments.

        else:
            result.append("<<Form%s(%s)>>" % (type, ",".join(
                adjustMacroArguments(parseMacroArguments(match), path, fragment, repeating, index)
                )))

    return "".join(result)

def adjustMacroArguments(args, path, fragment=None, repeating=None, index=None):

    """
    Adjust the given 'args' so that the path incorporates the given
    'path' and 'fragment', returning a new list containing the revised path,
    fragment and remaining arguments. If 'repeating' is specified, any field
    with such a name will be adjusted to reference the value with the given
    'index'.
    """

    if not path and not fragment and not repeating:
        return args

    result = []
    old_path = None
    found_name = None

    for arg in args:
        if arg.startswith("path="):
            old_path = arg[5:]
        elif arg.startswith("fragment=") and fragment:
            pass
        else:
            result.append(arg)
            if arg.startswith("name="):
                found_name = arg[5:]
            elif found_name is None:
                found_name = arg

    if path:
        qualified = old_path and ("%s/%s" % (old_path, path)) or path
        result.append("path=%s" % qualified)

    if fragment:
        result.append("fragment=%s" % fragment)

    if repeating and repeating == found_name:
        result.append("index=%s" % index)

    return result

def parseMacroArguments(args):

    """
    Interpret the arguments. To support commas in labels, the label argument
    should be quoted. For example:

    "label=No, thanks!"
    """

    try:
        parsed_args = args and wikiutil.parse_quoted_separated(args, name_value=False) or []
    except AttributeError:
        parsed_args = args.split(",")

    return [arg for arg in parsed_args if arg]

def getMacroArguments(parsed_args):

    "Return the macro arguments decoded from 'parsed_args'."

    name = None
    path = None
    dictpage = None
    label = None
    section = None
    fragment = None

    for arg in parsed_args:
        if arg.startswith("name="):
            name = arg[5:]

        elif arg.startswith("path="):
            path = arg[5:]

        elif arg.startswith("dict="):
            dictpage = arg[5:]

        elif arg.startswith("label="):
            label = arg[6:]

        elif arg.startswith("section="):
            section = arg[8:]

        elif arg.startswith("fragment="):
            fragment = arg[9:]

        elif name is None:
            name = arg

        elif dictpage is None:
            dictpage = arg

    return name, path, dictpage, label, section, fragment

def getFields(d, remove=False):

    """
    Return the form fields hierarchy for the given dictionary 'd'. If the
    optional 'remove' parameter is set to a true value, remove the entries for
    the fields from 'd'.
    """

    fields = {}

    for key, value in d.items():

        # Detect modifying fields.

        if key.find("=") != -1:
            fields[key] = value
            if remove:
                del d[key]
            continue

        # Reproduce the original hierarchy of the fields.

        section = fields
        parts = getPathDetails(key)

        for name, index in parts[:-1]:

            # Add an entry for instances of the section.

            if not section.has_key(name):
                section[name] = {}

            # Add an entry for the specific instance of the section.

            if not section[name].has_key(index):
                section[name][index] = {}

            section = section[name][index]

        section[parts[-1][0]] = value

        if remove:
            del d[key]

    return fields

def getPathDetails(path):

    """
    Return the given 'path' as a list of (name, index) tuples providing details
    of section instances, with any specific field appearing as the last element
    and having the form (name, None).
    """

    parts = []

    for part in path.split("/"):
        try:
            name, index = part.split("$", 1)
            index = int(index)
        except ValueError:
            name, index = part, None

        parts.append((name, index))

    return parts

def getSectionForPath(path, fields):

    """
    Obtain the section indicated by the given 'path' from the 'fields',
    returning a tuple of the form (parent section, (name, index)), where the
    parent section contains the referenced section, where name is the name of
    the referenced section, and where index, if not None, is the index of a
    specific section instance within the named section.
    """

    parts = getPathDetails(path)
    section = fields

    for name, index in parts[:-1]:
        section = fields[name][index]

    return section, parts[-1]

def getSectionElements(section_elements):

    "Return the given 'section_elements' as an ordered collection."

    keys = map(int, section_elements.keys())
    keys.sort()

    elements = []

    for key in keys:
        elements.append(section_elements[key])

    return elements

# Parser-related formatting functions.

def formatForm(text, request, fmt, attrs=None, write=None):

    """
    Format the given 'text' using the specified 'request' and formatter 'fmt'.
    The optional 'attrs' can be used to control the presentation of the form.

    If the 'write' parameter is specified, use it to write output; otherwise,
    write output using the request.
    """

    write = write or request.write
    page = request.page

    fields = getFields(get_form(request))

    # Prepare the query string for the form action URL.

    queryparams = []

    for argname, default in [("fragment", None), ("action", "MoinFormHandler")]:
        if attrs and attrs.has_key(argname):
            queryparams.append("%s=%s" % (argname, attrs[argname]))
        elif default:
            queryparams.append("%s=%s" % (argname, default))

    querystr = "&".join(queryparams)
    fragment = attrs.get("fragment")

    write(fmt.rawHTML('<form method="post" action="%s%s"%s>' % (
        escattr(page.url(request, querystr)),
        fragment and ("#%s" % escattr(fragment)) or "",
        fragment and (' id="%s"' % escattr(fragment)) or ""
        )))

    # Obtain page text for the form, incorporating subregions and applicable
    # sections.

    output = getFormOutput(text, fields, fragment=fragment)
    write(formatText(output, request, fmt, inhibit_p=False))

    write(fmt.rawHTML('</form>'))

def formatFormForOutputType(text, request, mimetype, attrs=None, write=None):

    """
    Format the given 'text' using the specified 'request' for the given output
    'mimetype'.

    The optional 'attrs' can be used to control the presentation of the form.

    If the 'write' parameter is specified, use it to write output; otherwise,
    write output using the request.
    """

    write = write or request.write

    if mimetype == "text/html":
        write('<html>')
        write('<body>')
        fmt = request.html_formatter
        fmt.setPage(request.page)
        formatForm(text, request, fmt, attrs, write)
        write('</body>')
        write('</html>')

# vim: tabstop=4 expandtab shiftwidth=4
